#region File Description
//-------------------------------------------------------------------------------------------------------
// File : AsuraOBJImporter.cs
// 
// Autodesk | Alias | Wavefront OBJ File Importer
// Date : Oct. 18, 2008
// Version : 1.0
// Author : Pocol
//-------------------------------------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.IO;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Globalization;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
#endregion

namespace Asura
{
	[ContentImporter(".obj", CacheImportedData = true, DefaultProcessor = "ModelProcessor", DisplayName ="OBJ File - Asura OBJ Importer")]
	public class AsuraOBJImporter : ContentImporter<NodeContent>
	{
		#region Variables
		ContentImporterContext importerContext;
		private NodeContent rootNode;
		private List<Vector3> positions;
		private List<Vector2> texcoords;
		private List<Vector3> normals;
		private MeshBuilder meshBuilder;
		private int[] positionMap;
		private int texcoordDataIndex;
		private int normalDataIndex;
		private Dictionary<String, MaterialContent> materials;
		private ContentIdentity mtlFileIdentity;
		private BasicMaterialContent materialContent;
		#endregion

		/// <summary>
		/// Yields an array of tokens for each line in an OBJ or MTL file.
		/// </summary>
		/// <remarks>
		/// OBJ and MTL files are text based formats of identical structure.
		/// Each line of a OBJ or MTL file is either an instruction or a comment.
		/// Comments begin with # and are effectively ignored.
		/// Instructions are a space dilimited list of tokens. The first token is the
		/// instruction type code. The tokens which follow, are the arguments to that
		/// instruction. The number and format of arguments vary by instruction type.
		/// </remarks>
		/// <param name="filename">Full path of file to read.</param>
		/// <param name="identity">Identity of the file being read. This is modified to
		/// include the current line number in case an exception is thrown.</param>
		/// <returns>Element 0 is the line type identifier. The remaining elements are
		/// arguments to the identifier's operation.</returns>
		private static IEnumerable<string[]> GetLineTokens(string filename, ContentIdentity identity)
		{
			using ( StreamReader reader = new StreamReader(filename) )
			{
				int lineNumber = 1;

				while ( !reader.EndOfStream)
				{
					identity.FragmentIdentifier = lineNumber.ToString();
					string[] lineTokens = Regex.Split(reader.ReadLine().Trim(), @"\s+");

					if ( lineTokens.Length > 0 && lineTokens[0] != String.Empty && !lineTokens[0].StartsWith("#"))
					{
						yield return lineTokens;
					}
					lineNumber++;
				}

				identity.FragmentIdentifier = null;
			}
		}

		/// <summary>
		/// Parse a Vector2 from tokens of an OBJ file line.
		/// </summary>
		/// <param name="lineTokens">X and Y coordinates in lineTokens[1 through 2].</param>
		private static Vector2 ParseVector2(string[] lineTokens)
		{
			return new Vector2(
				float.Parse(lineTokens[1], CultureInfo.InvariantCulture),
				float.Parse(lineTokens[2], CultureInfo.InvariantCulture)
			);
		}

		/// <summary>
		/// Parse a Vector3 from tokens of an OBJ file line.
		/// </summary>
		/// <param name="lineTokens">X, Y and Z coordinates in lineTokens[1 thorugh 3].</param>
		/// <returns></returns>
		private static Vector3 ParseVector3(string[] lineTokens)
		{
			return new Vector3(
				float.Parse(lineTokens[1], CultureInfo.InvariantCulture),
				float.Parse(lineTokens[2], CultureInfo.InvariantCulture),
				float.Parse(lineTokens[3], CultureInfo.InvariantCulture)
			);
		}

		/// <summary>
		/// The importer's entry point.
		/// Called by the framework when importing a game asset.
		/// </summary>
		/// <param name="filename">Name of a game asset file.</param>
		/// <param name="context">
		/// Contains information for importing a game asset, such as a logger interface.
		/// </param>
		/// <returns>Resulting game asset.</returns>
		public override NodeContent Import(string filename, ContentImporterContext context)
		{
			importerContext = context;

			rootNode = new NodeContent();
			positions = new List<Vector3>();
			texcoords = new List<Vector2>();
			normals = new List<Vector3>();
			meshBuilder = null;
			materials = new Dictionary<string, MaterialContent>();
			rootNode.Identity = new ContentIdentity(filename);

			try
			{
				foreach (String[] lineTokens in GetLineTokens(filename, rootNode.Identity))
				{
					ParseOBJLine(lineTokens);
				}

				if (rootNode.Name == null)
					rootNode.Name = Path.GetFileNameWithoutExtension(filename);


				FinishMesh();

				return rootNode;
			}
			catch (InvalidContentException)
			{
				throw;
			}
			catch (Exception e)
			{
				throw new InvalidContentException("Unable to parse obj file. Exception:\n" + e.Message, rootNode.Identity, e);
			}
		}

		/// <summary>
		/// Parses and executes an individual line of an OBJ file.
		/// </summary>
		/// <param name="lineTokens">Line to parse as tokens</param>
		private void ParseOBJLine(string[] lineTokens)
		{
			switch (lineTokens[0].ToLower())
			{
				//@IuWFNg
				case "o":
					{
						rootNode.Name = lineTokens[1];
						break;
					}

				//@_W
				case "v":
					{
						positions.Add(ParseVector3(lineTokens));
						break;
					}

				//@eNX`W
				case "vt":
					{
						Vector2 vt = ParseVector2(lineTokens);
						vt.Y = 1 - vt.Y;	//@VW𔽓]
						texcoords.Add(vt);
						break;
					}

				//@@xNg
				case "vn":
					{
						normals.Add(ParseVector3(lineTokens));
						break;
					}

				//@O[v
				case "g":
					{
						if (meshBuilder != null)
							FinishMesh();

						if (lineTokens.Length > 1)
							StartMesh(lineTokens[1]);
						else
							StartMesh(null);
						break;
					}

				//@X[WOO[v
				case "s":
					break;

				//@
				case "f":
					{
						// p`|S͔T|[g
						if (lineTokens.Length > 4)
						{
							importerContext.Logger.LogWarning(null, rootNode.Identity,
								   "N-sided polygons are not supported; Ignoring face");
							break;
						}
						if (meshBuilder == null)
							StartMesh(null);

						for (int vertexIndex = 1; vertexIndex <= 3; vertexIndex++)
						{
							string[] indices = lineTokens[vertexIndex].Split('/');

							int positionIndex = int.Parse(indices[0], CultureInfo.InvariantCulture) - 1;
							if (indices.Length > 1)
							{
								int texcoordIndex;
								Vector2 texcoord;
								if (int.TryParse(indices[1], out texcoordIndex))
									texcoord = texcoords[texcoordIndex - 1];
								else
									texcoord = Vector2.Zero;

								meshBuilder.SetVertexChannelData(texcoordDataIndex, texcoord);
							}

							if (indices.Length > 2)
							{
								int normalIndex;
								Vector3 normal;
								if (int.TryParse(indices[2], out normalIndex))
									normal = normals[normalIndex - 1];
								else
									normal = Vector3.Zero;

								meshBuilder.SetVertexChannelData(normalDataIndex, normal);
							}

							meshBuilder.AddTriangleVertex(positionMap[positionIndex]);
						}
						break;
					}

				//@}eAt@C
				case "mtllib":
					{
						for (int i = 1; i < lineTokens.Length; i++)
						{
							string mtlFileName = lineTokens[i];

							if (!Path.IsPathRooted(mtlFileName))
							{
								string directory = Path.GetDirectoryName(rootNode.Identity.SourceFilename);
								mtlFileName = Path.GetFullPath(Path.Combine(directory, mtlFileName));
							}

							importerContext.AddDependency(mtlFileName);

							ImportMaterials(mtlFileName);
						}
						break;
					}


				//@}eA
				case "usemtl":
					{
						if (meshBuilder == null)
							StartMesh(null);

						string materialName = lineTokens[1];

						MaterialContent material;
						if (materials.TryGetValue(materialName, out material))
							meshBuilder.SetMaterial(material);
						else
						{
							throw new InvalidContentException(String.Format("Material '{0}' ͒`Ă܂", materialName), rootNode.Identity);
						}
						break;
					}

				default:
					throw new InvalidContentException("Unsupported or invalid line type '" + lineTokens[0] + "'", rootNode.Identity);
			}

		}

		/// <summary>
		/// Starts a new mesh and fills it with mesh mapped positions.
		/// </summary>
		/// <param name="name">Name of mesh.</param>
		private void StartMesh(string name)
		{
			meshBuilder = MeshBuilder.StartMesh(name);

			texcoordDataIndex = meshBuilder.CreateVertexChannel<Vector2>(VertexChannelNames.TextureCoordinate(0));
			normalDataIndex = meshBuilder.CreateVertexChannel<Vector3>(VertexChannelNames.Normal());

			positionMap = new int[positions.Count];
			for (int i = 0; i < positions.Count; i++)
			{
				positionMap[i] = meshBuilder.CreatePosition(positions[i]);
			}
		}

		/// <summary>
		/// Finishes building a mesh and adds the resulting MeshContent or
		/// NodeContent to the root model's NodeContent.
		/// </summary>
		private void FinishMesh()
		{
			MeshContent meshContent = meshBuilder.FinishMesh();

			if (meshContent.Geometry.Count > 0)
			{
				rootNode.Children.Add(meshContent);
			}
			else
			{
				NodeContent nodeContent = new NodeContent();
				nodeContent.Name = meshContent.Name;

				rootNode.Children.Add(nodeContent);
			}
			meshBuilder = null;
		}

		/// <summary>
		/// Parses an MTL file and adds all its materials to the materials collection
		/// </summary>
		/// <param name="filename">Full path of MTL file.</param>
		private void ImportMaterials(string filename)
		{
			mtlFileIdentity = new ContentIdentity(filename);

			materialContent = null;

			try
			{
				foreach (String[] lineTokens in GetLineTokens(filename, mtlFileIdentity))
				{
					ParseMTLLine(lineTokens);
				}
			}
			catch (InvalidContentException)
			{
				throw;
			}
			catch (Exception e)
			{
				throw new InvalidContentException("Unable to parse mtl file. Exception:\n" + e.Message, mtlFileIdentity, e);
			}

			if (materialContent != null)
				materials.Add(materialContent.Name, materialContent);
		}

		/// <summary>
		/// Parses and executes an individual line of a MTL file.
		/// </summary>
		/// <param name="lineTokens">Line to parse as tokens</param>
		void ParseMTLLine(string[] lineTokens)
		{
			switch (lineTokens[0].ToLower())
			{
				//@V}eA̒`
				case "newmtl":
					{
						if (materialContent != null)
							materials.Add(materialContent.Name, materialContent);

						materialContent = new BasicMaterialContent();
						materialContent.Identity = new ContentIdentity(mtlFileIdentity.SourceFilename);

						materialContent.Name = lineTokens[1];
						break;
					}

				//@DiffuseJ[
				case "kd":
					{
						materialContent.DiffuseColor = ParseVector3(lineTokens);
						break;
					}

				//@Diffuse}bv
				case "map_kd":
					{
						materialContent.Texture = new ExternalReference<TextureContent>(lineTokens[1], mtlFileIdentity);
						break;
					}

				//@AmbientJ[
				case "ka":
					{
						materialContent.OpaqueData.Add("AmbientColor", ParseVector3(lineTokens));
						break;
					}

				//@SpecularJ[
				case "ks":
					{
						materialContent.SpecularColor = ParseVector3(lineTokens);
						break;
					}

				//@Specular Power
				case "ns":
					{
						materialContent.SpecularPower = float.Parse(lineTokens[1], CultureInfo.InvariantCulture);
						break;
					}

				//@EmissiveJ[
				case "ke":
					{
						materialContent.EmissiveColor = ParseVector3(lineTokens);
						break;
					}

				//@At@l
				case "d":
					{
						materialContent.Alpha = float.Parse(lineTokens[1], CultureInfo.InvariantCulture);
						break;
					}

				//@C~l[V[h
				case "illum":
					{
						materialContent.OpaqueData.Add("Illumination mode", int.Parse(lineTokens[1], CultureInfo.InvariantCulture));
						break;
					}

				default:
					throw new InvalidContentException(
						"Unsupported or invalid line type '" + lineTokens[0] + "'", mtlFileIdentity);
			}
		}
	}
}

